/*
 * pr_iface.c
 *
 * Protocol interface routines
 */

static char *rcsid = "/net/bcomp/src/master/nexus/pr_iface.c,v 1.62 1995/11/22 20:50:05 tuecke Exp";

#include "internal.h"

#include <netinet/in.h>
#ifdef NEXUS_ARCH_AXPOSF
#define NX_NTOHL(m) m
#define NX_HTONL(m) m
#else
#define NX_NTOHL(m) ntohl(m)
#define NX_HTONL(m) htonl(m)
#endif

#ifdef NEXUS_DEBUG
/* These are used by nexus_global_pointer_string() */
#define GP_N_STRINGS 10
static nexus_mutex_t gp_string_mutex;
#endif /* NEXUS_DEBUG */

#ifdef NEXUS_PROFILE
#define NEXUS_POLL_RANGE_CHECK
static nexus_bool_t check_poll_ranges;
static nexus_bool_t check_poll_traceback;
static double last_poll_time;
static double poll_range_min;
static double poll_range_max;
#endif


/*
 * mi_proto table stuff.
 *
 * The mi_proto table contains linked lists of nexus_mi_proto_t
 * structures.
 *
 * This table is used to avoid creating multiple nexus_mi_proto_t
 * objects that are the same.  Multiple global pointers which have
 * the same mi_proto information share a single nexus_mi_proto_t.
 */
#define MI_PROTO_TABLE_SIZE 1021
static nexus_mi_proto_t *	mi_proto_table[MI_PROTO_TABLE_SIZE];
static nexus_mutex_t		mi_proto_table_mutex;
static void			mi_proto_table_init(void);
static int			mi_proto_array_hash(nexus_byte_t *array,
						    int size);
static nexus_mi_proto_t *	mi_proto_table_insert(
					       nexus_mi_proto_t *new_mi_proto);

/*
 * Other typedefs
 */
static nexus_mi_proto_t *	mi_proto_create(int size,
						nexus_byte_t *array,
						nexus_proto_t *proto);
void				_nx_mi_proto_instantiate(
						nexus_mi_proto_t *mi_proto);
static int			same_gp(nexus_global_pointer_t *gp1,
					nexus_global_pointer_t *gp2,
					nexus_bool_t compare_addresses);

/*
 * Proto module list
 */
typedef struct _proto_module_list_t
{
    nexus_proto_funcs_t *		funcs;
    char *				name;
    nexus_proto_type_t			type;
    int					mi_proto_size;
    nexus_byte_t *			mi_proto_array;
    struct _proto_module_list_t *	next;
} proto_module_list_t;
static proto_module_list_t *	proto_module_list_head;
static proto_module_list_t *	proto_module_list_tail;

#define AddProtoModuleToList(Caller, Name, Funcs, Type) \
{ \
    proto_module_list_t *__p; \
    NexusMalloc(Caller, __p, proto_module_list_t *, \
		sizeof(proto_module_list_t)); \
    __p->name = _nx_copy_string(Name); \
    __p->funcs = (Funcs); \
    __p->type = (Type); \
    __p->mi_proto_size = 0; \
    __p->mi_proto_array = (nexus_byte_t *) NULL; \
    __p->next = (proto_module_list_t *) NULL; \
    if (proto_module_list_head) \
    { \
	proto_module_list_tail->next = __p; \
	proto_module_list_tail = __p; \
    } \
    else \
    { \
	proto_module_list_head = proto_module_list_tail = __p; \
    } \
}

static int			n_protos_with_poll;
static proto_module_list_t *	proto_module_with_poll;

static nexus_mi_proto_t *	my_mi_proto;
static nexus_proto_t *		local_proto;


/*
 * saved proto arguments
 */
static char saved_no_proto_arg[256];


/*
 * class_type_pointer_sizes
 *
 * There should be one entry in this table for each nexus_class_type_t.
 * This allows a machine to figure out how much of the liba is used
 * for the context and address.
 * The value for a given class_type should be the number of bytes
 * needed by that type of machine to encode a pointer.  (It is
 * probably the same size as unsigned long.)
 */
static int class_type_pointer_sizes[] =
{
    4, /* NEXUS_CLASS_TYPE_UNKNOWN */
    4, /* NEXUS_CLASS_TYPE_32BIT_LE */
    4, /* NEXUS_CLASS_TYPE_32BIT_BE */
    8, /* NEXUS_CLASS_TYPE_64BIT_LE */
    8, /* NEXUS_CLASS_TYPE_64BIT_BE */
    8, /* NEXUS_CLASS_TYPE_CRAYC90 */
    4, /* NEXUS_CLASS_TYPE_UNKNOWN32 */
    8, /* NEXUS_CLASS_TYPE_UNKNOWN64 */
};

/*
 * _nx_proto_usage_message()
 *
 * Call the usage_message() function for each protocol
 */
void _nx_proto_usage_message(void)
{
#ifdef NEXUS_POLL_RANGE_CHECK    
    printf("    -poll_check <min> <max>   : Print a notice if the time between\n");
    printf("                                nexus_poll() calls does not fall in the\n");
    printf("                                supplied range. (<min> and <max> are reals)\n");
    printf("    -poll_trace               : Print stack tracebacks with\n");
    printf("                                poll_check messages\n");
#endif /* NEXUS_POLL_RANGE_CHECK */	
} /* _nx_proto_usage_message() */


/*
 * _nx_proto_new_process_params()
 *
 * Call the new_process_params() function for each protocol
 *
 * Each of those functions may add stuff to 'buf', returning the number
 * of characters that they added.
 *
 * Return: The total number of characters added to 'buf'.
 */
int _nx_proto_new_process_params(char *buf, int size)
{
    char tmp_buf[256];
    int n_left = size;
    char *b = buf;
    int n_added;

    n_added = 0;
    tmp_buf[0] = '\0';

    if (strlen(saved_no_proto_arg) > 0)
    {
	char tmp_buf2[256];

	sprintf(tmp_buf2, "-no_pr %s ", saved_no_proto_arg);
	strcat(tmp_buf, tmp_buf2);
    }
#ifdef NEXUS_POLL_RANGE_CHECK
    if (check_poll_ranges)
    {
	char tmp_buf2[256];

	sprintf(tmp_buf2, "-poll_check %f %f ",
		poll_range_min, poll_range_max);
	if (check_poll_traceback)
	{
	    strcat(tmp_buf2, "-poll_trace ");
	}
	strcat(tmp_buf, tmp_buf2);
    }
#endif /* NEXUS_POLL_RANGE_CHECK */	
    n_added = strlen(tmp_buf);
    if (n_added > n_left)
    {
        nexus_fatal("_nx_proto_new_process_params(): Internal error: Not enough room in buffer for arguments\n");
    }
    strcpy(b, tmp_buf);
    b += n_added;
    n_left -= n_added;

    return (size - n_left);
} /* _nx_proto_new_process_params() */


/*
 * _nx_proto_init()
 *
 * Initialize the protocol modules.
 *
 * TODO: This needs to be made more dynamic
 */
void _nx_proto_init(int *argc, char ***argv,
		    nexus_module_list_t module_list[])
{
    int i;
    int rc;
    nexus_proto_funcs_t *proto_funcs;
    nexus_proto_type_t proto_type;
    proto_module_list_t *proto_module;
    nexus_byte_t *array;
    int size;
    int node_name_length;
    int arg_num;
    char *arg;
    char *next, *no_proto;
    nexus_bool_t add_proto;
    int pid;

    /*
     * Parse any arguments
     */
    if ((arg_num = nexus_find_argument(argc, argv, "no_pr", 2)) >= 0)
    {
	arg = (*argv)[arg_num + 1];
	strcpy(saved_no_proto_arg, arg);
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    
#ifdef NEXUS_POLL_RANGE_CHECK
    check_poll_ranges = NEXUS_FALSE;
    check_poll_traceback = NEXUS_FALSE;
    last_poll_time = -1.0;
    if ((arg_num = nexus_find_argument(argc, argv, "poll_check", 3)) >= 0)
    {
	check_poll_ranges = NEXUS_TRUE;
	poll_range_min = atof((*argv)[arg_num + 1]);
	poll_range_max = atof((*argv)[arg_num + 2]);
	nexus_remove_arguments(argc, argv, arg_num, 3);
    }
    if ((arg_num = nexus_find_argument(argc, argv, "poll_trace", 1)) >= 0)
    {
	check_poll_traceback = NEXUS_TRUE;
	nexus_remove_arguments(argc, argv, arg_num, 1);
    }
#endif /* NEXUS_POLL_RANGE_CHECK */

#ifdef NEXUS_DEBUG
    nexus_mutex_init(&gp_string_mutex, (nexus_mutexattr_t *) NULL);
#endif

    /*
     * Scan the module_list looking for protocol modules.
     * For each of these, get the function table,
     * and add that module to the proto_module_list.
     * Put the local protocol module at the front of the list.
     */
    proto_module_list_head = proto_module_list_tail
	= (proto_module_list_t *) NULL;
    proto_funcs = _nx_pr_local_info();
    proto_type = (*proto_funcs->proto_type)();
    AddProtoModuleToList(_nx_proto_init(),
			 "local",
			 proto_funcs,
			 proto_type);

    for (i = 0; module_list[i].family_name != (char *) NULL; i++)
    {
	if ((strcmp(module_list[i].family_name, "protocols") == 0))
	{
	    add_proto = NEXUS_TRUE;
	    next = saved_no_proto_arg;
	    /* 
	     * This can be changed to while(next && add_proto) if the
	     * number of no_pr options increase.  With only tcp, atm,
	     * and inx, this is not an optimization.
	     */
	    while (next)
	    {
		_nx_get_next_value(next, ':', &next, &no_proto);
		if (strcmp(module_list[i].module_name, no_proto) == 0)
		{
		    add_proto = NEXUS_FALSE;
		}
	    }
	    /*
	    rc = _nx_module_load(module_list[i].family_name,
				 module_list[i].module_name,
				 (void *) &proto_funcs);
	    */
	    rc = 0;
	    if (rc == 0 && add_proto)
	    {
	        proto_funcs
		    = (nexus_proto_funcs_t *) (*module_list[i].info_func)();
		proto_type = (*proto_funcs->proto_type)();
		AddProtoModuleToList(_nx_proto_init(),
				     module_list[i].module_name,
				     proto_funcs,
				     proto_type);
	    }
	}
    }

    /*
     * Initialize each protocol module
     */
    for (proto_module = proto_module_list_head;
	 proto_module;
	 proto_module = proto_module->next)
    {
	(*proto_module->funcs->init)(argc, argv);
    }

    /*
     * Find out how many of the protos have poll functions.
     */
    n_protos_with_poll = 0;
    for (proto_module = proto_module_list_head;
	 proto_module;
	 proto_module = proto_module->next)
    {
	if (proto_module->funcs->poll)
	{
	    n_protos_with_poll++;
	    proto_module_with_poll = proto_module;
	}
    }

    /*
     * Get the nexus_proto_t for the local protocol module
     * and fill it into my_mi_proto.
     */
    (proto_module_list_head->funcs->construct_from_mi_proto)(&local_proto,
						(nexus_mi_proto_t *) NULL,
						(nexus_byte_t *) NULL,
						0);

    /*
     * Construct my nexus_mi_proto_t.
     * First ask each protocol module what size its component will be,
     * then malloc an array big enough, and then ask
     * then ask each protocol module to fill their component into
     * the array.
     */
    mi_proto_table_init();
    node_name_length = strlen(_nx_my_node.name);
    size = 8 + node_name_length + 1;
    for (proto_module = proto_module_list_head;
	 proto_module;
	 proto_module = proto_module->next)
    {
	if (proto_module->funcs->get_my_mi_proto)
	{
	    (proto_module->funcs->get_my_mi_proto)(
					      &(proto_module->mi_proto_array),
					      &(proto_module->mi_proto_size) );
	    size += 4 + proto_module->mi_proto_size;
	}
    }
    my_mi_proto = mi_proto_create(size, (nexus_byte_t *) NULL, local_proto);
    array = my_mi_proto->array;
    i = 0;
    PackInt4(array, i, _nx_my_node.number);
    pid = _nx_md_getpid();
    PackInt4(array, i, pid);
    memcpy(&(array[i]), _nx_my_node.name, node_name_length + 1);
    i += node_name_length + 1;
    for (proto_module = proto_module_list_head;
	 proto_module;
	 proto_module = proto_module->next)
    {
	if (proto_module->mi_proto_size > 0)
	{
	    PackMIProtoEntry(array, i,
			     proto_module->type,
			     proto_module->mi_proto_size,
			     proto_module->mi_proto_array);
	}
    }
    mi_proto_table_insert(my_mi_proto);

} /* _nx_proto_init() */


/*
 * _nx_proto_shutdown()
 *
 * If 'shutdown_others' is NEXUS_TRUE,
 * then try to shutdown all other nodes and contexts.
 */
void _nx_proto_shutdown(nexus_bool_t shutdown_others)
{
    proto_module_list_t *proto_module;
    
    for (proto_module = proto_module_list_head;
	 proto_module;
	 proto_module = proto_module->next)
    {
	if (proto_module->funcs->shutdown)
	{
	    (proto_module->funcs->shutdown)(shutdown_others);
	}
    }
} /* _nx_proto_shutdown() */


/*
 * _nx_proto_abort()
 */
void _nx_proto_abort(void)
{
    proto_module_list_t *proto_module;
    
    for (proto_module = proto_module_list_head;
	 proto_module;
	 proto_module = proto_module->next)
    {
	if (proto_module->funcs->abort)
	{
	    (proto_module->funcs->abort)();
	}
    }
} /* _nx_proto_abort() */


/*
 * nexus_poll
 *
 * Poll each proto, so that they can handle any outstanding remote
 * service requests.
 *
 * No need to poll the local protocol.
 */
void nexus_poll(void)
{
    proto_module_list_t *proto_module;
    
    for (proto_module = proto_module_list_head;
	 proto_module;
	 proto_module = proto_module->next)
    {
	if (proto_module->funcs->poll)
	{
	    (proto_module->funcs->poll)();
	}
    }
} /* nexus_poll() */


/*
 * nexus_poll_profile
 *
 * Poll each proto, so that they can handle any outstanding remote
 * service requests.
 *
 * No need to poll the local protocol.
 */
void nexus_poll_profile(int id)
{
    
#ifdef NEXUS_POLL_RANGE_CHECK
    if (check_poll_ranges && last_poll_time >= 0)
    {
	double time_since_last_poll;
	time_since_last_poll = nexus_wallclock() - last_poll_time;
	if (time_since_last_poll < poll_range_min)
	{
	    nexus_notice("poll id %d: Time since last nexus_poll() (%f) is less than minimum\n",
			 id, time_since_last_poll);
	    if (check_poll_traceback)
	    {
		_nx_traceback();
	    }
	}
	else if (time_since_last_poll > poll_range_max)
	{
	    nexus_notice("poll id %d: Time since last nexus_poll() (%f) exceeds maximum\n",
			 id, time_since_last_poll);
	    if (check_poll_traceback)
	    {
		_nx_traceback();
	    }
	}
    }
#endif /* NEXUS_POLL_RANGE_CHECK */    

    nexus_poll();
    
#ifdef NEXUS_POLL_RANGE_CHECK
    if (check_poll_ranges)
    {
	last_poll_time = nexus_wallclock();
    }
#endif /* NEXUS_POLL_RANGE_CHECK */    
    
} /* nexus_poll_profile() */


/*
 * nexus_blocking_poll()
 *
 * If only one proto has a poll routine, then call
 * its blocking_poll() routine.
 * If more than one proto has a poll routine, then call
 * each of those poll() routines.
 *
 * No need to poll the local protocol.
 */
void nexus_blocking_poll(void)
{
    if (n_protos_with_poll == 1)
    {
	(*proto_module_with_poll->funcs->blocking_poll)();
    }
    else if (n_protos_with_poll > 1)
    {
	nexus_poll();
    }
} /* nexus_blocking_poll() */


/*
 * _nx_must_poll()
 *
 * Return NEXUS_TRUE if nexus_poll() and/or nexus_blocking_poll()
 * should be used -- that is,
 * are there one or more protocol modules with poll routines.
 * Return NEXUS_FALSE otherwise.
 */
nexus_bool_t _nx_must_poll(void)
{
    if (n_protos_with_poll >= 1)
	return (NEXUS_TRUE);
    else
	return (NEXUS_FALSE);
} /* _nx_must_poll() */


/*
 * nexus_global_pointer()
 */
void nexus_global_pointer(nexus_global_pointer_t *new_gp,
			  void *address)
{
    nexus_context_t *context;
    _nx_context(&context);

    /* Ignore the reference count on local gps */
    new_gp->mi_proto = my_mi_proto;
    
    new_gp->class_type = NEXUS_CLASS_TYPE;
#ifdef NEXUS_PROFILE
    _nx_node_id(&(new_gp->node_id));
    _nx_context_id(&(new_gp->context_id));
#endif /* NEXUS_PROFILE */
    new_gp->liba_size = 2 * sizeof(unsigned long);
    if (new_gp->liba_size <= NEXUS_DEFAULT_LIBA_SIZE)
    {
	new_gp->liba_is_inline = 1;
	PackLiba(new_gp->liba.array,
		 (unsigned long) context,
		 (unsigned long) address);
    }
    else
    {
	new_gp->liba_is_inline = 0;
	NexusMalloc(nexus_global_pointer(),
		    new_gp->liba.pointer,
		    nexus_byte_t *,
		    new_gp->liba_size);
	PackLiba(new_gp->liba.pointer,
		 (unsigned long) context,
		 (unsigned long) address);
    }
} /* nexus_global_pointer() */


/*
 * nexus_convert_global_pointer_address()
 */
void *nexus_convert_global_pointer_address(nexus_global_pointer_t *gp)
{
    unsigned long c, a;
    NexusAssert2((nexus_global_pointer_to_current_context(gp)),
		 ("nexus_convert_global_pointer_address(): Global pointer does not point to this context\n") );
    UnpackLibaFromGP(gp, c, a);
    return ((void *) a);
} /* nexus_convert_global_pointer_address() */


/*
 * nexus_destroy_global_pointer()
 */
void nexus_destroy_global_pointer(nexus_global_pointer_t *gp)
{
    nexus_mi_proto_t *mi_proto;
    nexus_proto_t *proto;

    if (!nexus_macro_is_null_global_pointer(gp))
    {
	mi_proto = gp->mi_proto;
	proto = mi_proto->proto;
	if (proto->funcs->decrement_reference_count)
	{
	    (proto->funcs->decrement_reference_count)(proto);
	}
	
	if (!gp->liba_is_inline)
	{
	    NexusFree(gp->liba.pointer);
	}
	
	nexus_macro_null_global_pointer(gp);
    }
} /* nexus_destroy_global_pointer() */


/*
 * nexus_copy_global_pointer()
 *
 * Copies the global pointer gp_in to gp_out allowing for
 * reference counting in the protocol modules.
 */
void nexus_copy_global_pointer(nexus_global_pointer_t *gp_out,
			       nexus_global_pointer_t *gp_in)
{
    gp_out->mi_proto = gp_in->mi_proto;
    gp_out->class_type = gp_in->class_type;
#ifdef NEXUS_PROFILE
    gp_out->node_id = gp_in->node_id;
    gp_out->context_id = gp_in->context_id;
#endif /* NEXUS_PROFILE */

    gp_out->liba_is_inline = gp_in->liba_is_inline;
    gp_out->liba_size = gp_in->liba_size;
    if (gp_in->liba_is_inline)
    {
	memcpy(gp_out->liba.array, gp_in->liba.array, gp_in->liba_size);
    }
    else
    {
	NexusMalloc(nexus_copy_global_pointer(),
		    gp_out->liba.pointer,
		    nexus_byte_t *,
		    gp_in->liba_size);
	memcpy(gp_out->liba.pointer, gp_in->liba.pointer, gp_in->liba_size);
    }
    
    if (gp_in->mi_proto->proto->funcs->increment_reference_count)
    {
	(gp_in->mi_proto->proto->funcs->increment_reference_count)(
						      gp_in->mi_proto->proto);
    }
} /* nexus_copy_global_pointer() */


/*
 * nexus_sizeof_global_pointer()
 *
 * Return the size in bytes needed to encode the passed
 * global pointer, 'gp', into the 'buffer'.
 * Also, set 'n_elements' to the number of puts that will be
 * done when nexus_put_global_pointer() is called for this gp.
 */
int nexus_sizeof_global_pointer(nexus_buffer_t *buffer,
				nexus_global_pointer_t *gp,
				int count,
				int *n_elements)
{
    nexus_global_pointer_t *g;
    int c;
    int size = 0;
    int n = 0;

    NexusAssert2((gp != (nexus_global_pointer_t *) NULL),
		 ("nexus_sizeof_global_pointer(): Passed a NULL pointer to a gp.\n"));
		
    for (c = 0, g = gp;
	 c < count;
	 c++, g++)
    {
	/* sizeof class_type */
	size += nexus_sizeof_int(buffer, 1);
	n++;

	if (!nexus_macro_is_null_global_pointer(g))
	{
#ifdef NEXUS_PROFILE
	    /* sizeof the node_id, context_id */
	    size += nexus_sizeof_int(buffer, 1);
	    size += nexus_sizeof_int(buffer, 1);
	    n += 2;
#endif /* NEXUS_PROFILE */    

	    /* sizeof liba */
	    size += nexus_sizeof_int(buffer, 1);
	    size += nexus_sizeof_byte(buffer, g->liba_size);
	    n += 2;
	    
	    /* sizeof mi_proto */
	    size += nexus_sizeof_int(buffer, 1);
	    size += nexus_sizeof_byte(buffer, g->mi_proto->size);
	    n += 2;
	}
    }
    
    if (n_elements != (int *) NULL)
    {
	*n_elements = n;
    }
    
    return (size);
} /* nexus_sizeof_global_pointer() */


/*
 * nexus_put_global_pointer()
 */
void nexus_put_global_pointer(nexus_buffer_t *buffer,
			      nexus_global_pointer_t *gp,
			      int count)
{
    nexus_global_pointer_t *g;
    int c;
    int tmp_int;

    for (c = 0, g = gp;
	 c < count;
	 c++, g++)
    {
	/* check that everything is ok first */
	NEXUS_INTERROGATE(gp,_NX_GLOBAL_POINTER_T,"nexus_put_global_pointer");
	
	if (nexus_macro_is_null_global_pointer(g))
	{
	    /* class_type==0 means its a null gp */
	    tmp_int = 0;
	    nexus_put_int(buffer, &tmp_int, 1);
	}
	else
	{
	    /* put the class_type */
	    tmp_int = (int) g->class_type;
	    nexus_put_int(buffer, &tmp_int, 1);
	
#ifdef NEXUS_PROFILE
	    /* put the node_id, context_id */
	    nexus_put_int(buffer, &(g->node_id), 1);
	    nexus_put_int(buffer, &(g->context_id), 1);
#endif /* NEXUS_PROFILE */    

	    /* put the liba */
	    tmp_int = g->liba_size;
	    nexus_put_int(buffer, &tmp_int, 1);
	    if (g->liba_is_inline)
	    {
		nexus_put_byte(buffer, (void *) g->liba.array, g->liba_size);
	    }
	    else
	    {
		nexus_put_byte(buffer, (void *) g->liba.pointer, g->liba_size);
	    }
	    
	    /* put the mi_proto */
	    nexus_put_int(buffer, &(g->mi_proto->size), 1);
	    nexus_put_byte(buffer, g->mi_proto->array, g->mi_proto->size);
	}
    }
} /* nexus_put_global_pointer() */


/*
 * nexus_get_global_pointer()
 */
void nexus_get_global_pointer(nexus_buffer_t *buffer,
			      nexus_global_pointer_t *gp,
			      int count)
{
    nexus_global_pointer_t *g;
    int c;
    int tmp_int;
    int mi_proto_size;
    nexus_mi_proto_t *mi_proto;

    for (c = 0, g = gp;
	 c < count;
	 c++, g++)
    {
	/* get the class_type */
	nexus_get_int(buffer, &tmp_int, 1);

	if (tmp_int == 0)
	{
	    /* class_type==0 means its a null gp */
	    nexus_macro_null_global_pointer(g);
	}
	else
	{
	    g->class_type = (nexus_class_type_t) tmp_int;
	
#ifdef NEXUS_PROFILE
	    /* get the node_id, context_id */
	    nexus_get_int(buffer, &(g->node_id), 1);
	    nexus_get_int(buffer, &(g->context_id), 1);
#endif /* NEXUS_PROFILE */    

	    /* get the liba */
	    nexus_get_int(buffer, &tmp_int, 1);
	    g->liba_size = tmp_int;
	    if (g->liba_size <= NEXUS_DEFAULT_LIBA_SIZE)
	    {
		g->liba_is_inline = 1;
		nexus_get_byte(buffer,
			       (void *) g->liba.array,
			       g->liba_size);
	    }
	    else
	    {
		g->liba_is_inline = 0;
		NexusMalloc(nexus_get_global_pointer(),
			    g->liba.pointer,
			    nexus_byte_t *,
			    g->liba_size);
		nexus_get_byte(buffer,
			       (void *) g->liba.pointer,
			       g->liba_size);
	    }
	    
	    /* get the mi_proto */
	    nexus_get_int(buffer, &mi_proto_size, 1);
	    mi_proto = mi_proto_create(mi_proto_size,
				       (nexus_byte_t *) NULL,
				       (nexus_proto_t *) NULL);
	    nexus_get_byte(buffer, mi_proto->array, mi_proto_size);
	    g->mi_proto = mi_proto_table_insert(mi_proto);
	    
	    NEXUS_INTERROGATE(g,
			      _NX_GLOBAL_POINTER_T,
			      "nexus_get_global_pointer");
	}
    }
	
} /* nexus_get_global_pointer() */


/*
 * nexus_get_stashed_global_pointer()
 *
 * This is identical to nexus_get_global_pointer(), except that
 * nexus_get_stashed_TYPE() is used instead of nexus_get_TYPE().
 */
void nexus_get_stashed_global_pointer(nexus_stashed_buffer_t *buffer,
				      nexus_global_pointer_t *gp,
				      int count)
{
    nexus_global_pointer_t *g;
    int c;
    int tmp_int;
    int mi_proto_size;
    nexus_mi_proto_t *mi_proto;

    for (c = 0, g = gp;
	 c < count;
	 c++, g++)
    {
	g->mi_proto = (nexus_mi_proto_t *) NULL;

	/* get the class_type */
	nexus_get_stashed_int(buffer, &tmp_int, 1);

	if (tmp_int == 0)
	{
	    /* class_type==0 means its a null gp */
	    nexus_macro_null_global_pointer(g);
	}
	else
	{
	    g->class_type = (nexus_class_type_t) tmp_int;
	
#ifdef NEXUS_PROFILE
	    /* get the node_id, context_id */
	    nexus_get_stashed_int(buffer, &(g->node_id), 1);
	    nexus_get_stashed_int(buffer, &(g->context_id), 1);
#endif /* NEXUS_PROFILE */    

	    /* get the liba */
	    nexus_get_stashed_int(buffer, &tmp_int, 1);
	    g->liba_size = tmp_int;
	    if (g->liba_size <= NEXUS_DEFAULT_LIBA_SIZE)
	    {
		g->liba_is_inline = 1;
		nexus_get_stashed_byte(buffer,
				       (void *) g->liba.array,
				       g->liba_size);
	    }
	    else
	    {
		g->liba_is_inline = 0;
		NexusMalloc(nexus_get_stashed_global_pointer(),
			    g->liba.pointer,
			    nexus_byte_t *,
			    g->liba_size);
		nexus_get_stashed_byte(buffer,
				       (void *) g->liba.pointer,
				       g->liba_size);
	    }

	    /* get the mi_proto */
	    nexus_get_stashed_int(buffer, &mi_proto_size, 1);
	    mi_proto = mi_proto_create(mi_proto_size,
				       (nexus_byte_t *) NULL,
				       (nexus_proto_t *) NULL);
	    nexus_get_stashed_byte(buffer, mi_proto->array, mi_proto_size);
	    g->mi_proto = mi_proto_table_insert(mi_proto);
	    
	    NEXUS_INTERROGATE(g,
			      _NX_GLOBAL_POINTER_T,
			      "nexus_get_stashed_global_pointer");
	}
    }
	
} /* nexus_get_stashed_global_pointer() */


/*
 * _nx_write_global_pointer()
 *
 * Write the global pointer to the passed file descriptor.
 *
 * Return:
 *	0 on success
 *	1 if the filedesc is unexpectedly closed
 */
int _nx_write_global_pointer(int filedesc, nexus_global_pointer_t *gp)
{
    int tmp_int;

    NEXUS_INTERROGATE(gp, _NX_GLOBAL_POINTER_T, "_nx_write_global_pointer");

    if (nexus_macro_is_null_global_pointer(gp))
    {
	/* class_type==0 means its a null gp */
	tmp_int = 0;
	if (_nx_write_int4(filedesc, tmp_int) == 0)
	    return(1);
    }
    else
    {
	/* put class_type */
	tmp_int = (int) gp->class_type;
	if (_nx_write_int4(filedesc, tmp_int) == 0)
	    return(1);
	
#ifdef NEXUS_PROFILE
	/* put the node_id, context_id */
	if (_nx_write_int4(filedesc, gp->node_id) == 0)
	    return(1);
	if (_nx_write_int4(filedesc, gp->context_id) == 0)
	    return(1);
#endif /* NEXUS_PROFILE */    
	
	/* put the liba */
	tmp_int = gp->liba_size;
	if (_nx_write_int4(filedesc, tmp_int) == 0)
	    return(1);
	if (gp->liba_is_inline)
	{
	    if (_nx_write(filedesc,
			  (char *) gp->liba.array,
			  gp->liba_size) == 0)
		return(1);
	}
	else
	{
	    if (_nx_write(filedesc,
			  (char *) gp->liba.pointer,
			  gp->liba_size) == 0)
		return(1);
	}
	
	/* put the mi_proto */
	if (_nx_write_int4(filedesc, gp->mi_proto->size) == 0)
	    return(1);
	if (_nx_write(filedesc,
		      (char *) gp->mi_proto->array,
		      gp->mi_proto->size)==0)
	    return(1);
    }

    return(0);
    
} /* _nx_write_global_pointer() */


/*
 * _nx_read_global_pointer()
 *
 * Read a global pointer from the passed file descriptor and fill
 * it into 'gp'.
 *
 * Return:
 *	0 on success
 *	1 if the filedesc is unexpectedly closed
 */
int _nx_read_global_pointer(int filedesc, nexus_global_pointer_t *gp)
{
    int tmp_int;
    int mi_proto_size;
    nexus_mi_proto_t *mi_proto;

    /* get the context, address */
    if (_nx_read_int4(filedesc, &tmp_int) == 0)
	return(1);

    if (tmp_int == 0)
    {
	/* class_type==0 means its a null gp */
	nexus_macro_null_global_pointer(gp);
    }
    else
    {
	gp->class_type = (nexus_class_type_t) tmp_int;

#ifdef NEXUS_PROFILE
	/* get the node_id, context_id */
	if (_nx_read_int4(filedesc, &(gp->node_id)) == 0)
	    return(1);
	if (_nx_read_int4(filedesc, &(gp->context_id)) == 0)
	    return(1);
#endif /* NEXUS_PROFILE */    
	
	/* get the liba */
	if (_nx_read_int4(filedesc, &tmp_int) == 0)
	    return(1);
	gp->liba_size = tmp_int;
	if (gp->liba_size <= NEXUS_DEFAULT_LIBA_SIZE)
	{
	    gp->liba_is_inline = 1;
	    if (_nx_read(filedesc,
			 (char *) gp->liba.array,
			 gp->liba_size) == 0)
		return(1);
	}
	else
	{
	    gp->liba_is_inline = 0;
	    NexusMalloc(nexus_get_global_pointer(),
			gp->liba.pointer,
			nexus_byte_t *,
			gp->liba_size);
	    if (_nx_read(filedesc,
			 (char *) gp->liba.pointer,
			 gp->liba_size) == 0)
		return(1);
	}
	
	/* get the mi_proto */
	if (_nx_read_int4(filedesc, &mi_proto_size) == 0)
	    return(1);
	mi_proto = mi_proto_create(mi_proto_size,
				   (nexus_byte_t *) NULL,
				   (nexus_proto_t *) NULL);
	if (_nx_read(filedesc, (char *) mi_proto->array, mi_proto_size) == 0)
	    return(1);
	gp->mi_proto = mi_proto_table_insert(mi_proto);
    }
	
    NEXUS_INTERROGATE(gp, _NX_GLOBAL_POINTER_T, "_nx_read_global_pointer");

    return(0);
    
} /* _nx_read_global_pointer() */


#ifdef NEXUS_DEBUG
/*
 * nexus_global_pointer_string()
 *
 * Return a string representation of the global pointer.
 * The returned string does not need to be freed, coming as it does
 * from a pool of static buffers. Note we lock in here.
 */
char *nexus_global_pointer_string(nexus_global_pointer_t *gp)
{
    static char strings[GP_N_STRINGS][1024];
    static int current_string = 0;
    char *buf;
    unsigned long context, address;

    nexus_mutex_lock(&gp_string_mutex);
    buf = strings[current_string];
    current_string = ((current_string + 1) % GP_N_STRINGS);
    nexus_mutex_unlock(&gp_string_mutex);
    
    UnpackLibaFromGP(gp, context, address);
    
    buf[0] = '\0';

#ifdef NEXUS_PROFILE
    sprintf(buf, "[GP:%lx:%lx:%lx:n%d:c%d]",
	    (unsigned long) gp,
	    context,
	    address,
	    gp->node_id,
	    gp->context_id);
#else  /* NEXUS_PROFILE */    
    sprintf(buf, "[GP:%lx:%lx:%lx]",
	    (unsigned long) gp,
	    context,
	    address);
#endif /* NEXUS_PROFILE */    
              
    return(buf);
} /* nexus_global_pointer_string() */
#endif /* NEXUS_DEBUG */


/*
 * same_gp()
 *
 * Compare two gp's to see if they are in the same context.
 * If compare_addresses==NEXUS_TRUE, then also compare if
 * they point to the same address.
 *
 * Return NEXUS_TRUE if they are same, or NEXUS_FALSE if they are different.
 */
static int same_gp(nexus_global_pointer_t *gp1,
		   nexus_global_pointer_t *gp2,
		   nexus_bool_t compare_addresses)
{
    nexus_byte_t *a1, *a2;
    int size;
    
    NEXUS_INTERROGATE(gp1, _NX_GLOBAL_POINTER_T, "same_gp");
    NEXUS_INTERROGATE(gp2, _NX_GLOBAL_POINTER_T, "same_gp");

    if (   (gp1->mi_proto == (nexus_mi_proto_t *) NULL)
	&& (gp2->mi_proto == (nexus_mi_proto_t *) NULL) )
    {
	/* Both are NULL global pointers */
	return (NEXUS_TRUE);
    }
    if (   (gp1->mi_proto == (nexus_mi_proto_t *) NULL)
	|| (gp2->mi_proto == (nexus_mi_proto_t *) NULL) )
    {
	/* One is a NULL global pointer, the other isn't */
	return (NEXUS_FALSE);
    }

    /* Compare their class_types */
    if (gp1->class_type != gp2->class_type)
	return (NEXUS_FALSE);
    
    /* Compare their liba's */
    if (gp1->liba_is_inline)
	a1 = gp1->liba.array;
    else
	a1 = gp1->liba.pointer;
    if (gp2->liba_is_inline)
	a2 = gp2->liba.array;
    else
	a2 = gp2->liba.pointer;
    size = class_type_pointer_sizes[gp1->class_type];
    if (compare_addresses)
	size *= 2;
    if (memcmp(a1, a2, size) != 0)
	return (NEXUS_FALSE);

    /* Compare their node numbers and context differentiators */
    a1 = gp1->mi_proto->array;
    a2 = gp2->mi_proto->array;
    if (memcmp(a1, a2, 8) != 0)
	return (NEXUS_FALSE);

    /* Compare their node names */
    if (strcmp((char *) (a1+4), (char *) (a2+4)) != 0)
	return (NEXUS_FALSE);
    
    /* They are the same context */
    return (NEXUS_TRUE);
} /* same_gp() */


/*
 * nexus_same_context()
 */
int nexus_same_context(nexus_global_pointer_t *gp1,
		       nexus_global_pointer_t *gp2)
{
    return (same_gp(gp1, gp2, NEXUS_FALSE));
} /* nexus_same_context() */


/*
 * nexus_same_global_pointer()
 */
int nexus_same_global_pointer(nexus_global_pointer_t *gp1,
			      nexus_global_pointer_t *gp2)
{
    return (same_gp(gp1, gp2, NEXUS_TRUE));
} /* nexus_same_global_pointer() */


/*
 * nexus_global_pointer_to_current_context()
 */
int nexus_global_pointer_to_current_context(nexus_global_pointer_t *gp)
{
    nexus_global_pointer_t gp_to_myself;
    int rc;
    nexus_global_pointer(&gp_to_myself, NULL);
    rc = same_gp(gp, &gp_to_myself, NEXUS_FALSE);
    nexus_destroy_global_pointer(&gp_to_myself);
    return(rc);
} /* nexus_global_pointer_to_current_context() */


/*
 * _nx_proto_get_creator_gp_params()
 *
 * Fill in 'buf', up to 'buf_size' bytes, with a command line
 * argument encoding * of the passed global pointer, 'gp', so
 * that a new process can connect to this one.
 */
void _nx_proto_get_creator_gp_params(char *buf,
				     int buf_size,
				     nexus_global_pointer_t *gp)
{
    int len;
    char tmp_buf[64];
#ifdef NEXUS_PROFILE
    int my_node_id, my_context_id;
#endif /* NEXUS_PROFILE */

    /*
     * Put parameters into 'buf' to specify:
     *   -nf_nx_gp <class_type>			global pointer info
     *   -nf_nx_gp_id <context_id>:<node_id>	id's for profiling
     *   -nf_nx_gp_liba <hex_encoded_liba>	global pointer's liba
     *   -nf_nx_gp_mi <hex_encoded_mi_proto>	global pointer's mi_proto
     */

#ifdef NEXUS_PROFILE
    _nx_node_id(&my_node_id);
    _nx_context_id(&my_context_id);
    nexus_stdio_lock();
    sprintf(tmp_buf, "-nf_nx_gp_id :%d:%d ", my_context_id, my_node_id);
    nexus_stdio_unlock();
#else  /* NEXUS_PROFILE */
    tmp_buf[0] = '\0';
#endif /* NEXUS_PROFILE */

    if ((size_t) buf_size < (64  /* strings + fudging */
			     + strlen(tmp_buf)
			     + (2*gp->liba_size)
			     + (2*gp->mi_proto->size) ) )
    {
	nexus_fatal("_nx_proto_get_creator_gp_params(): Internal error: Buffer not big enough for arguments\n");
    }

    nexus_stdio_lock();
    sprintf(buf, "-nf_nx_gp %d %s-nf_nx_gp_liba ",
	    gp->class_type,
	    tmp_buf);
    nexus_stdio_unlock();
    len = strlen(buf);
    _nx_hex_encode_byte_array((gp->liba_is_inline
			       ? gp->liba.array
			       : gp->liba.pointer),
			      gp->liba_size,
			      &(buf[len]) );
    len += (2 * gp->liba_size);
    memcpy(&(buf[len]), " -nf_nx_gp_mi ", 14);	/* without null terminator */
    len += 14;
    _nx_hex_encode_byte_array(gp->mi_proto->array,
			      gp->mi_proto->size,
			      &(buf[len]) );
    len += (2 * gp->mi_proto->size);
    memcpy(&(buf[len]), " ", 2);		/* with null terminator */
	
} /* _nx_proto_get_creator_gp_params */


/*
 * _nx_proto_construct_creator_gp()
 */
void _nx_proto_construct_creator_gp(nexus_global_pointer_t *creator_gp)
{
    char *string;
    nexus_mi_proto_t *mi_proto;
    int tmp_int;
    int size;
    int rc;
    int arg_num;
    int *argc;
    char ***argv;

    nexus_get_retained_arguments(&argc, &argv);
    
    if ((arg_num = nexus_find_argument(argc, argv, "nf_nx_gp", 2)) >= 0)
    {
	string = (*argv)[arg_num + 1];
	nexus_remove_arguments(argc, argv, arg_num, 2);
	
	nexus_stdio_lock();
	rc = sscanf(string, "%d", &tmp_int);
	nexus_stdio_unlock();
	if (rc != 1)
	{
	    nexus_fatal("_nx_proto_construct_creator_gp(): Internal error: Invalid -nf_nx_gp argument.\n");
	}
	
	creator_gp->class_type = (nexus_class_type_t) tmp_int;
    }
    else
    {
	nexus_fatal("_nx_proto_construct_creator_gp(): Internal error: Expected a -nf_nx_gp argument.\n");
    }
    
    if ((arg_num = nexus_find_argument(argc, argv, "nf_nx_gp_liba", 2)) >= 0)
    {
	string = (*argv)[arg_num + 1];
	nexus_remove_arguments(argc, argv, arg_num, 2);

	creator_gp->liba_size = strlen(string) / 2;
	if (creator_gp->liba_size <= NEXUS_DEFAULT_LIBA_SIZE)
	{
	    creator_gp->liba_is_inline = 1;
	    _nx_hex_decode_byte_array(string,
				      creator_gp->liba_size,
				      creator_gp->liba.array);
	}
	else
	{
	    creator_gp->liba_is_inline = 0;
	    NexusMalloc(_nx_proto_construct_creator_gp(),
			creator_gp->liba.pointer,
			nexus_byte_t *,
			creator_gp->liba_size);
	    _nx_hex_decode_byte_array(string,
				      creator_gp->liba_size,
				      creator_gp->liba.pointer);
	}
    }
    else
    {
	nexus_fatal("_nx_proto_construct_creator_gp(): Internal error: Expected a -nf_nx_gp_liba argument.\n");
    }

    if ((arg_num = nexus_find_argument(argc, argv, "nf_nx_gp_mi", 2)) >= 0)
    {
	string = (*argv)[arg_num + 1];
	nexus_remove_arguments(argc, argv, arg_num, 2);

	size = strlen(string) / 2;
	mi_proto = mi_proto_create(size,
				   (nexus_byte_t *) NULL,
				   (nexus_proto_t *) NULL);
	mi_proto->size = size;
	_nx_hex_decode_byte_array(string,
				  mi_proto->size,
				  mi_proto->array);
	creator_gp->mi_proto = mi_proto_table_insert(mi_proto);
    }
    else
    {
	nexus_fatal("_nx_proto_construct_creator_gp(): Internal error: Expected a -nf_nx_gp_liba argument.\n");
    }

#ifdef NEXUS_PROFILE    
    /*
     * Get the global pointer context and address
     */
    if ((arg_num = nexus_find_argument(argc, argv, "nf_nx_gp_id", 2)) >= 0)
    {
	string = (*argv)[arg_num + 1];
	nexus_remove_arguments(argc, argv, arg_num, 2);
	
	nexus_stdio_lock();
	rc = sscanf(string, ":%d:%d",
		    &(creator_gp->context_id),
		    &(creator_gp->node_id));
	nexus_stdio_unlock();
	if (rc != 2)
	{
	    nexus_fatal("_nx_proto_construct_creator_gp(): Internal error: Invalid -nf_nx_gp_id argument.\n");
	}
    }
    else
    {
	creator_gp->context_id = -1;
	creator_gp->node_id = -1;
	nexus_debug_printf(1, ("_nx_proto_construct_creator_gp(): No -nf_nx_gp_id argument.\n"));
    }
#endif /* NEXUS_PROFILE */
    
} /* _nx_proto_construct_creator_gp() */


/*
 * nexus_test_global_pointer()
 *
 * Return 0 if the passed global pointer is (probably) still working,
 * else return non-0.
 *
 * If the protocol module for this gp does not implement the test_proto
 * function, then return 0.
 */
int nexus_test_global_pointer(nexus_global_pointer_t *gp)
{
    int rc;
    if (   (gp)
	&& (gp->mi_proto)
	&& (gp->mi_proto->proto->funcs->test_proto))
    {
	rc = (gp->mi_proto->proto->funcs->test_proto)(gp->mi_proto->proto);
    }
    else
    {
	rc = 0;
    }
    return(rc);
} /* nexus_test_global_pointer() */


/*
 * _nx_write()
 *
 * This is a wrapper around write() which will keep trying until
 * 'len' characters have been written.  This will work on either
 * blocking or non-blocking file descriptors, and will properly
 * retry on EINTR.
 *
 * Return:
 *	'len' on success
 *	0 if the fd is closed before 'len' bytes are read
 */
int _nx_write(int fd, char *buf, int len)
{
    char *b = buf;
    int l = len;
    int rc;
    int write_errno;

    while (l > 0)
    {
	rc = write(fd, b, l);
	if (rc >= 0)
	{
	    b += rc;
	    l -= rc;
	}
	else
	{
	    write_errno = errno;
	    if (write_errno == EINTR || write_errno == EWOULDBLOCK)
	    {
		/* Do nothing.  Just try again. */
	    }
	    else if (write_errno == EPIPE || write_errno == ECONNRESET)
	    {
		/* fd was closed */
		nexus_debug_printf(1, ("_nx_write(): fd was unexpectedly closed\n"));
		return(0);
	    }
	    else
	    {
		nexus_fatal("_nx_write(): Write failed: %s\n",
			    _nx_md_system_error_string(write_errno));
	    }
	}
    }
    return(len);
} /* _nx_write() */


/*
 * _nx_read()
 *
 * This is a wrapper around read() which will keep trying until
 * 'len' characters have been read.  This will work on either
 * blocking or non-blocking file descriptors, and will properly
 * retry on EINTR.
 *
 * Return:
 *	'len' on success
 *	0 if EOF happens on the fd before 'len' bytes are read
 */
int _nx_read(int fd, char *buf, int len)
{
    char *b = buf;
    int l = len;
    int rc;
    int read_errno;

    while (l > 0)
    {
	rc = read(fd, b, l);
	read_errno = errno;
	if (rc > 0)
	{
	    b += rc;
	    l -= rc;
	}
	else if (   (rc == 0)
		 || (   (rc < 0)
		     && (   (read_errno == ECONNRESET)
			 || (read_errno == EPIPE) ) ) )
	{
	    nexus_debug_printf(1, ("_nx_read(): fd unexpectedly closed: errno=%d: %s\n", read_errno, _nx_md_system_error_string(read_errno)));
	    return(0);
	}
	else /* rc < 0 */
	{
	    if (read_errno == EINTR || read_errno== EAGAIN ||
		read_errno == EWOULDBLOCK)
	    {
		/* Do nothing.  Just try again. */
	    }
	    else
	    {
		nexus_fatal("_nx_read(): Read failed: %s\n",
			    _nx_md_system_error_string(read_errno));
	    }
	}
    }
    return(len);
} /* _nx_read() */


/*
 * _nx_write_int4()
 *
 * Write the 'i' to 'fd' as a 4 byte integer.
 *
 * Return:
 *	non-0 on success
 *	0 if EOF happens on the fd before 'i' can be written
 */
int _nx_write_int4(int fd, int i)
{
    nexus_byte_t buf[4];
    int j = 0;
    PackInt4(buf, j, i);
    return(_nx_write(fd, (char *) buf, 4));
} /* _nx_write_int4() */


/*
 * _nx_read_int4()
 *
 * Read the 4 byte integer '*i' from 'fd'
 *
 * Return:
 *	non-0 on success
 *	0 if EOF happens on the fd before 'i' can be written
 */
int _nx_read_int4(int fd, int *i)
{
    nexus_byte_t buf[4];
    int rc;
    int j = 0;
    rc = _nx_read(fd, (char *) buf, 4);
    UnpackInt4(buf, j, *i);
    return(rc);
} /* _nx_read_int4() */


/*
 * mi_proto_create()
 *
 * Return an mi_proto for the passed 'size, 'array', and 'proto'.
 * If array==NULL, then do not try to copy it.  It will be filled in later.
 */
static nexus_mi_proto_t *mi_proto_create(int size,
					 nexus_byte_t *array,
					 nexus_proto_t *proto)
{
    nexus_mi_proto_t *new_mi_proto;

    NexusMalloc(mi_proto_create(),
		new_mi_proto,
		nexus_mi_proto_t *,
		(sizeof(nexus_mi_proto_t) + size));
    new_mi_proto->proto = proto;
    new_mi_proto->next = (nexus_mi_proto_t *) NULL;
    new_mi_proto->size = size;
    new_mi_proto->array = (nexus_byte_t *) (((char *) new_mi_proto)
					    + sizeof(nexus_mi_proto_t));
    if (array)
    {
	memcpy(new_mi_proto->array, array, size);
    }
    
    return (new_mi_proto);
    
} /* mi_proto_create() */


/*
 * _nx_mi_proto_instantiate()
 *
 * Go through the entries in the mi_proto's array, looking for ones
 * that we can use.  When one is found, instantiate it into
 * a nexus_proto_t object.
 */
void _nx_mi_proto_instantiate(nexus_mi_proto_t *mi_proto)
{
    proto_module_list_t *proto_module;
    nexus_proto_t *proto;
    nexus_byte_t *a;
    nexus_byte_t *proto_array;
    int i;
    nexus_proto_type_t type;
    int size;
    nexus_bool_t done;
    int node_number;
    int context_differentiator;
    char *node_name;

    a = mi_proto->array;
    i = 0;

    NexusAssert2(((i + 9) <= mi_proto->size),
		 ("_nx_mi_proto_instantiate(): Invalid mi_proto array.  Not enough room for header.\n"));
    
    UnpackMIProtoHeader(a, i,
			node_number,
			context_differentiator,
			node_name);
    
    done = NEXUS_FALSE;
    while (!done && (i < mi_proto->size))
    {
	NexusAssert2(((i + 4) <= mi_proto->size),
		     ("_nx_mi_proto_instantiate(): Invalid mi_proto array.  Not enough room for protocol entry header.\n"));

	UnpackMIProtoEntry(a, i, type, size, proto_array);

	NexusAssert2((i <= mi_proto->size),
		     ("_nx_mi_proto_instantiate(): Invalid mi_proto array.  Not enough room for protocol entry.\n"));

	/* See if this protocol is in my proto_module_list */
	for (proto_module = proto_module_list_head;
	     !done && proto_module;
	     proto_module = proto_module->next)
	{
	    if (proto_module->type == type)
	    {
		/* Found a type match.  So call the proto to instantiate it */
		if ((proto_module->funcs->construct_from_mi_proto)(&proto,
								   mi_proto,
								   proto_array,
								   size))
		{
		    /* Successful instantiation */
		    if (!proto)
		    {
			/* Use local_proto */
			proto = local_proto;
		    }
		    mi_proto->proto = proto;
		    done = NEXUS_TRUE;
		}
	    }
	}
    }
    
    NexusAssert2((mi_proto->proto != (nexus_proto_t *) NULL),
		 ("_nx_mi_proto_instantiate(): Could not find a usable protocol.\n"));

} /* _nx_mi_proto_instantiate() */


/*
 * mi_proto_table_init()
 *
 * Initialize the mi_proto hash table.
 */
static void mi_proto_table_init(void)
{
    int i;
    for (i = 0; i < MI_PROTO_TABLE_SIZE; i++)
    {
	mi_proto_table[i] = (nexus_mi_proto_t *) NULL;
    }
    nexus_mutex_init(&mi_proto_table_mutex, (nexus_mutexattr_t *) NULL);
} /* mi_proto_table_init() */


/*
 * mi_proto_array_hash()
 *
 * Hash the passed mi_proto 'array'.
 */
static int mi_proto_array_hash(nexus_byte_t *array, int size)
{
    unsigned long hval = 0;
    int i;

    for (i = 0; i < size; i++)
    {
	hval <<= 1;
	hval += (unsigned long) array[i];
    }
    return ((int) (hval % MI_PROTO_TABLE_SIZE));
} /* mi_proto_array_hash() */


/*
 * mi_proto_table_insert()
 *
 * If 'new_mi_proto' already exists in the mi_proto_table,
 * then free it and return the existing copy.
 * Otherwise, add new_mi_proto to the table and return it.
 */
static nexus_mi_proto_t *mi_proto_table_insert(nexus_mi_proto_t *new_mi_proto)
{
    int bucket;
    nexus_mi_proto_t *mi_proto;

    bucket = mi_proto_array_hash(new_mi_proto->array, new_mi_proto->size);

    /* Search the bucket for this mi_proto */
    for (mi_proto = mi_proto_table[bucket];
	 mi_proto != (nexus_mi_proto_t *) NULL;
	 mi_proto = mi_proto->next)
    {
	if (   (new_mi_proto->size == mi_proto->size)
	    && (memcmp(new_mi_proto->array, mi_proto->array,
		       new_mi_proto->size) == 0) )
	{
	    /*
	     * Found mi_proto with this array in the table.
	     * So use the existing one, and free the new one.
	     * Bump the reference count of the proto for the existing one.
	     */
	    NexusFree(new_mi_proto);
	    if (mi_proto->proto->funcs->increment_reference_count)
	    {
		(mi_proto->proto->funcs->increment_reference_count)(
							mi_proto->proto);
	    }
	    return(mi_proto);
	}
    }
    
    /*
     * We did not find new_mi_proto in the table.
     * So add this new one to this hash bucket.
     */
    new_mi_proto->next = mi_proto_table[bucket];
    mi_proto_table[bucket] = new_mi_proto;

    if (new_mi_proto->proto == (nexus_proto_t *) NULL)
    {
	_nx_mi_proto_instantiate(new_mi_proto);
    }

    return (new_mi_proto);
    
} /* mi_proto_table_insert() */


