/*
 * process.c
 *
 * Code to handle process startup.
 */

static char *rcsid = "/net/bcomp/src/master/nexus/process.c,v 1.57 1995/11/13 22:51:56 tuecke Exp";

#include "internal.h"
#include <fcntl.h>
#ifdef NEXUS_ARCH_SOLARIS
#include <netdb.h>
#endif

nexus_bool_t		_nx_node_should_exit = NEXUS_FALSE;

static char *		arg_node_name;
static int		arg_node_number;
static nexus_bool_t	arg_use_node_name;
static int		arg_node_id;
static int              arg_num_nodes = 0;
static int		arg_context_size;
static int		arg_checkin_number;

static nexus_node_t     *other_nodes = NULL;
static int              n_other_nodes = 0;

#ifdef NEXUS_PROFILE
typedef struct _wait_for_node_id_t
{
    int *		node_id;
    nexus_bool_t	done;
    nexus_mutex_t	mutex;
    nexus_cond_t	cond;
} wait_for_node_id_t;

static int			next_node_id;
static nexus_mutex_t		next_node_id_mutex;

static void get_next_node_id(int *node_id, int n_nodes);

#define _NX_GET_NEXT_NODE_ID_HANDLER_HASH 886
static void _nx_get_next_node_id_handler(void *address,
					 nexus_buffer_t *buffer);

#define _NX_GET_NEXT_NODE_ID_REPLY_HANDLER_HASH 516
static void _nx_get_next_node_id_reply_handler(void *address,
					       nexus_buffer_t *buffer);
#endif /* NEXUS_PROFILE */

#ifdef DONT_INCLUDE
#define _NX_NODE_DEFER_HANDLER_HASH 242
static void _nx_node_defer_handler(void *address,
				   nexus_buffer_t *buffer);
#endif /* DONT_INCLUDE */

static nexus_handler_t process_handlers[] =
{
#ifdef DONT_INCLUDE
  {"_nx_node_defer_handler",
       _NX_NODE_DEFER_HANDLER_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _nx_node_defer_handler},
#endif /* DONT_INCLUDE */
#ifdef NEXUS_PROFILE
  {"_nx_get_next_node_id_handler",
       _NX_GET_NEXT_NODE_ID_HANDLER_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _nx_get_next_node_id_handler},
  {"_nx_get_next_node_id_reply_handler",
       _NX_GET_NEXT_NODE_ID_REPLY_HANDLER_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) _nx_get_next_node_id_reply_handler},
#endif /* NEXUS_PROFILE */
  {(char *) NULL,
       0,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) NULL},
};


/*
 * _nx_process_usage_message()
 */
void _nx_process_usage_message(void)
{
} /* _nx_process_usage_message() */


/*
 * _nx_process_new_process_params()
 */
int _nx_process_new_process_params(char *buf, int size)
{
    char tmp_buf1[MAXHOSTNAMELEN + 32];
    int len;
    
    sprintf(tmp_buf1, "-nx_mid %s ", _nx_master_id_string);
    len = strlen(tmp_buf1);
    if (len > size)
    {
	nexus_fatal("_nx_process_new_process_params(): Internal error: Not enough room in buffer for arguments\n");
    }
    strcpy(buf, tmp_buf1);
    return (len);
} /* _nx_process_new_process_params() */


/*
 * _nx_process_register_handlers()
 */
void _nx_process_register_handlers(void)
{
    nexus_register_handlers(process_handlers);
} /* _nx_process_register_handlers() */


/*
 * nexus_master_node()
 */
int nexus_master_node(void)
{
    return (_nx_master_node);
} /* nexus_master_node() */


/*
 * _nx_process_init()
 *
 * Initialize this new process.  This process may be the master
 * node, a non-master node, or a new context.
 *
 * Create appropriate node and context structures for this process.
 *
 * If I am the master node, create other nodes (as requested via
 * parameters).  Those other nodes will subsequently reach this
 * same point.
 *
 * Return:
 *	On the master node:
 * 		In '*nodes', an array of nexus_node_t's
 *			to the other nodes.
 *		In '*n_nodes', the number of nodes in 'nodes'.
 *	On other processes:
 *		In '*nodes', return NULL.
 *		In '*n_nodes', return 0.
 */
void _nx_process_init(int *argc,
		      char ***argv,
		      char *package_designator,
		      int (*package_args_init_func)(int *argc, char **argv[]),
		      nexus_module_list_t module_list[],
		      nexus_node_t **nodes,
		      int *n_nodes)
{
    nexus_context_t *context = (nexus_context_t *) NULL;
    char hostname[MAXHOSTNAMELEN];
    int arg_num;
    int context_size;
#ifdef NEXUS_PROFILE
    int node_id;
    int context_id;
    int thread_id;
#endif /* NEXUS_PROFILE */

    /*
     * Initialize the fundamental modules.
     */
    ports0_init(argc, argv, package_designator);
    _nx_args_init(argc, argv);
    if (package_args_init_func)
    {
	int rc;
	if ((rc = (package_args_init_func)(argc, argv)) != 0)
	{
	    _nx_md_exit(rc);
	}
    }
    _nx_md_init(argc, argv);
#ifndef NEXUS_LITE
    _nx_thread_init(argc, argv);
#endif /* NEXUS_LITE */
    _nx_context_init(argc, argv);
    _nx_handler_init(argc, argv);
    _nx_database_init(argc, argv, module_list);
    _nx_fault_tolerance_init(argc, argv);
    _nx_startup_init(argc, argv);
    
    /*
     * Get any of this module's arguments
     */
    arg_node_name = (char *) NULL;
    arg_node_number = -1;
    arg_use_node_name = NEXUS_FALSE;
    arg_node_id = -1;
    arg_context_size = -1;
    arg_checkin_number = -1;
    _nx_master_id_string = (char *) NULL;
    if ((arg_num = nexus_find_argument(argc, argv, "nx_node", 4)) >= 0)
    {
	arg_node_name = (*argv)[arg_num + 1];
	arg_node_number = atoi((*argv)[arg_num + 2]);
	arg_use_node_name = (*((*argv)[arg_num + 3]) == 'y');
	nexus_remove_arguments(argc, argv, arg_num, 4);
    }
    if ((arg_num = nexus_find_argument(argc, argv, "nx_node_id", 2)) >= 0)
    {
	arg_node_id = atoi((*argv)[arg_num + 1]);
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    if ((arg_num = nexus_find_argument(argc, argv, "nx_num_nodes", 2)) >= 0)
    {
	arg_num_nodes = atoi((*argv)[arg_num + 1]);
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    if ((arg_num = nexus_find_argument(argc, argv, "nx_context", 2)) >= 0)
    {
	arg_context_size = atoi((*argv)[arg_num + 1]);
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    if ((arg_num = nexus_find_argument(argc, argv, "nx_checkin", 2)) >= 0)
    {
	arg_checkin_number = atoi((*argv)[arg_num + 1]);
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    if ((arg_num = nexus_find_argument(argc, argv, "nx_mid", 2)) >= 0)
    {
	_nx_master_id_string = _nx_copy_string((*argv)[arg_num + 1]);
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    
#ifdef NEXUS_PROFILE
    next_node_id = -1;
    nexus_mutex_init(&next_node_id_mutex, (nexus_mutexattr_t *) NULL);
#endif /* NEXUS_PROFILE */

    if (arg_node_name == (char *) NULL)
    {
	/*
	 * This process is the master node.
	 */
	_nx_master_node = NEXUS_TRUE;
	_nx_my_node_id = 0;
	_nx_my_node.number = 0;
	_nx_my_node.return_code = NEXUS_NODE_NEW;
	_nx_my_node.name = "";
	_nx_my_node.name = _nx_startup_get_master_node_name();
	_nx_master_id_string = _nx_md_get_unique_session_string();
	
	context = _nx_context_alloc(0, NEXUS_TRUE);

	_nx_process_is_context = NEXUS_FALSE;
    }
    else
    {
	/*
	 * From the arguments we can tell that this 
	 * process is for a non-master node or a context:
	 *	- A -nx_node argument is present, so this is not
	 *	  the master node.
	 *	- If the process is for a context, then there
	 *	  will also be a -nx_context argument.
	 */
	_nx_master_node = NEXUS_FALSE;
	_nx_my_node_id = arg_node_id;
	_nx_my_node.return_code = NEXUS_NODE_NEW;
	_nx_my_node.number = arg_node_number;
	if (arg_use_node_name)
	{
	    /*
	     * Use arg_node_name as the name of this node
	     */
	    _nx_my_node.name = _nx_copy_string(arg_node_name);
	}
	else
	{
	    /*
	     * Ignore the arg_node_name, and use
	     * gethostname() instead, so as to get a canonical
	     * name for this node.
	     */
	    _nx_my_node.name = "";
	    _nx_md_gethostname(hostname, MAXHOSTNAMELEN);
	    _nx_my_node.name = _nx_copy_string(hostname);
	}

	if (arg_context_size >= 0)
	{
	    /* This is a context */
	    _nx_process_is_context = NEXUS_TRUE;
	    context_size = arg_context_size;
	}
	else
	{
	    /*
	     * This is a non-master node
	     */
	    _nx_process_is_context = NEXUS_FALSE;
	    context_size = 0;
	}

	if (_nx_master_id_string == (char *) NULL)
	{
	    printf("Fatal error: _nx_process_init(): Internal error: Expected a -nx_mid argument\n");
	    nexus_abort();
	}
	
	/*
	 * Initialize my context structure.
	 * Do this even for a non-master node, just to have
	 * someplace to put the handlers, etc.
	 */
	context = _nx_context_alloc(context_size, !_nx_process_is_context);
    }
    
    /* Set the thread's context */
    _nx_set_context(context);
    _nx_initial_context = context;

    /*
     * Initialize the other modules.
     * We want to wait until now to do this so that the
     * thread and context structures are all in place.
     * This allows, for example, protocol modules to create threads.
     */
#ifdef NEXUS_PROFILE    
    _nx_pablo_init(argc, argv);
    _nx_node_id(&node_id);
    _nx_context_id(&context_id);
    _nx_thread_id(&thread_id);
    if (!_nx_process_is_context)
    {
	_nx_pablo_log_node_creation(node_id);
    }
    _nx_pablo_log_context_creation(node_id, context_id);
    _nx_pablo_log_thread_creation(node_id, context_id, thread_id);
#endif

    _nx_proto_init(argc, argv, module_list);

#ifndef NEXUS_LITE
    _nx_thread_create_handler_thread();
#endif /* NEXUS_LITE */
    
    nexus_debug_printf(2, ("_nx_process_init(): _nx_master_node=%d\n", 
			   _nx_master_node));

    if (!_nx_process_is_context)
    {
	/* Filling the global pointer in _nx_my_node */
	nexus_global_pointer(&(_nx_my_node.gp), NULL);
    }

    if (_nx_master_node)
    {
	_nx_nodelock_master_init();
	
#ifdef NEXUS_PROFILE	
	/*
	 * Now that everything is initialize, setup the master_gp
	 */
	nexus_global_pointer(&(context->master_gp), NULL);
#endif	

    	/* Create the rest of the initial nodes */
	_nx_startup_initial_nodes(&_nx_my_node, nodes, n_nodes);

#ifdef NEXUS_PROFILE	
	next_node_id = *n_nodes;
#endif
    }
    else
    {
	if (arg_num_nodes >= 1)
	{
	    _nx_startup_initial_nodes(&_nx_my_node,
				      &other_nodes,
				      &n_other_nodes);
	}
	else
	{
	    other_nodes = &_nx_my_node;
	    n_other_nodes = 1;
	}
	*nodes = (nexus_node_t *) NULL;
	*n_nodes = 0;
    }

} /* _nx_process_init() */


/*
 * _nx_process_start()
 *
 * Call NexusBoot() if this process is the master node
 * or if it is for a new context.
 *
 * Then, if this process is not the master node, connect the
 * process back to its creator.
 */
void _nx_process_start(void)
{
    int nexus_boot_rc = 0;
  
    if (_nx_master_node)
    {
	nexus_boot_rc = (*_nx_NexusBoot)();
    }
    else /* (!_nx_master_node) */
    {
	/*
	 * Checkin with the process that created this one
	 */
	nexus_global_pointer_t creator_gp;
	void *initial_segment;
	
	/* Get my gp for returning in the checkin */
	initial_segment = nexus_context_initial_segment();
	nexus_global_pointer(&(_nx_my_node.gp), initial_segment);
	
	/* Construct a gp to the creator process */
	_nx_proto_construct_creator_gp(&creator_gp);
	
	NexusAssert2((arg_checkin_number >= 0),
		     ("_nx_process_process_arguments(): Internal error: Got a negative checkin number (-nx_checkin) argument.\n"));
	
	/* Invoke the appropriate checkin handler */
	if (_nx_process_is_context)
	{
	    nexus_boot_rc = (*_nx_NexusBoot)();
	    _nx_context_checkin(&(_nx_my_node.gp),
				nexus_boot_rc,
				&creator_gp,
				arg_checkin_number
#ifdef NEXUS_PROFILE
				, NEXUS_TRUE
#endif			    
				);
	}
	else
	{
	    _nx_node_should_exit = _nx_nodelock_check();
	    if (!_nx_node_should_exit)
	    {
		nexus_boot_rc = (*_nx_NexusBoot)();
	    }

	    _nx_node_checkin(arg_num_nodes >= 1 ? other_nodes : &_nx_my_node,
			     arg_num_nodes >= 1 ? n_other_nodes : 1,
			     &creator_gp,
			     arg_checkin_number
#ifdef NEXUS_PROFILE
			     , NEXUS_TRUE
#endif			    
			     );
	}
	
	nexus_destroy_global_pointer(&(_nx_my_node.gp));
	nexus_destroy_global_pointer(&creator_gp);
    }

    /*
     * Exit the process if:
     *      NexusBoot() returns non-zero
     *   or There is already a node processes here with my node name.
     */
    if (nexus_boot_rc > 0 || _nx_node_should_exit)
    {
	nexus_debug_printf(1, ("exiting because of non-zero nexus_boot_rc or node is deferring\n"));
	_nx_exit_transient_process(0);
    }
} /* _nx_process_start() */


#ifdef NEXUS_PROFILE
/*
 * get_next_node_id()
 *
 * Set *node_id to be the next available node id,
 * and increment next_node_id by 'n_nodes'.
 */
static void get_next_node_id(int *node_id, int n_nodes)
{
    nexus_mutex_lock(&next_node_id_mutex);
    *node_id = next_node_id;
    next_node_id += n_nodes;
    nexus_mutex_unlock(&next_node_id_mutex);
} /* get_next_node_id() */


/*
 * _nx_get_next_node_id_from_master()
 *
 * Set *node_id to be the next available node id,
 * and increment next_node_id by 'n_nodes'.
 *
 * We have to go to the master (context->master_gp) to get the node_id.
 */
void _nx_get_next_node_id_from_master(int *node_id, int n_nodes,
				      nexus_context_t *context)
{
    wait_for_node_id_t wait;
    nexus_global_pointer_t wait_gp;
    nexus_buffer_t buffer;
    int gp_size, gp_n_elements;

    /*
     * Wait for the master_gp to be filled in.
     * If it is NULL, it means the message containing the master_gp
     * is still in transit.
     */
    if (nexus_is_null_global_pointer(&(context->master_gp)))
    {
	_nx_wait_for_master_gp_from_creator(context);
    }
    
    /* Initialize wait structure */
    wait.node_id = node_id;
    wait.done = NEXUS_FALSE;
    nexus_mutex_init(&(wait.mutex), (nexus_mutexattr_t *) NULL);
    nexus_cond_init(&(wait.cond), (nexus_condattr_t *) NULL);

    /* Send message to master node */
    nexus_global_pointer(&wait_gp, (void *) &wait);
    nexus_init_remote_service_request(&buffer,
				      &(context->master_gp),
				      "_nx_get_next_node_id_handler",
				      _NX_GET_NEXT_NODE_ID_HANDLER_HASH);
    gp_size = nexus_sizeof_global_pointer(&buffer, &wait_gp, 1,
					  &gp_n_elements);
    nexus_set_buffer_size(&buffer,
			  (gp_size + nexus_sizeof_int(&buffer, 1)),
			  (gp_n_elements + 1) );
    nexus_put_int(&buffer, &n_nodes, 1);
    nexus_put_global_pointer(&buffer, &wait_gp, 1);
    nexus_send_remote_service_request(&buffer);

    /* Wait for reply */
    nexus_mutex_lock(&(wait.mutex));
    while (!wait.done)
    {
	nexus_cond_wait(&(wait.cond), &(wait.mutex));
    }
    nexus_mutex_unlock(&(wait.mutex));

    nexus_destroy_global_pointer(&wait_gp);
    nexus_mutex_destroy(&(wait.mutex));
    nexus_cond_destroy(&(wait.cond));
    
} /* _nx_get_next_node_id_from_master() */


/*
 * _nx_get_next_node_id_handler()
 */
static void _nx_get_next_node_id_handler(void *address, nexus_buffer_t *buffer)
{
    int node_id;
    int n_nodes;
    nexus_global_pointer_t wait_gp;
    nexus_buffer_t reply_buffer;

    nexus_get_int(buffer, &n_nodes, 1);
    nexus_get_global_pointer(buffer, &wait_gp, 1);

    get_next_node_id(&node_id, n_nodes);
    
    /* Send reply message */
    nexus_init_remote_service_request(&reply_buffer,
				      &wait_gp,
				      "_nx_get_next_node_id_reply_handler",
				      _NX_GET_NEXT_NODE_ID_REPLY_HANDLER_HASH);
    nexus_set_buffer_size(&reply_buffer,
			  nexus_sizeof_int(&reply_buffer, 1),
			  1 );
    nexus_put_int(&reply_buffer, &node_id, 1);
    nexus_send_remote_service_request(&reply_buffer);

    nexus_destroy_global_pointer(&wait_gp);
    
} /* _nx_get_next_node_id_handler() */


/*
 * _nx_get_next_node_id_reply_handler()
 */
static void _nx_get_next_node_id_reply_handler(void *address,
					       nexus_buffer_t *buffer)
{
    wait_for_node_id_t *wait = (wait_for_node_id_t *) address;
    int node_id;

    nexus_get_int(buffer, &node_id, 1);

    *(wait->node_id) = node_id;
    
    /* Signal completion of reply */
    nexus_mutex_lock(&(wait->mutex));
    wait->done = NEXUS_TRUE;
    nexus_cond_signal(&(wait->cond));
    nexus_mutex_unlock(&(wait->mutex));
    
} /* _nx_get_next_node_id_reply_handler() */

#endif /* NEXUS_PROFILE */
