/*
 * st_start.h
 *
 * Code to connect to a secure server, talk to the secure server
 * to start a process, and then redirect stdout from that new process
 * to my stdout.
 *
 * This is a header file because the code is used in two ways:
 *   1) listener.c is an independent program which uses this code.
 *	This is needed by multi-threaded code where single-thread-fork
 *	semantics are broken.
 *   2) st_ss.c calls this code directly in a non-multi-threaded environment.
 *
 * This code is derived from p4.
 *
 * static char *rcsid = "/net/bcomp/src/master/nexus/ss_start.h,v 1.6 1995/11/20 20:09:34 tuecke Exp";
 */

/*
 * Function prototypes
 */
static char * start_secure_remote_node(char *host,
				       int port,
				       char *username,
				       char *dir,
				       char **env,
				       int n_env,
				       char *cmd,
				       char *args,
				       char *type);
static int connect_to_server(char *host, int port);
static int send_integer(int sock, int value);
static int send_string(int sock, char *str);
static int recv_string(int sock, char *str, int len);
static int setup_anon_listener(int backlog, int *port, int *sock);
static void getpw_ss(char *host, char *username,
		     char *buf, int buf_length);
static int echo(void);

#ifndef SS_Malloc
#define SS_Malloc(Size) malloc(Size)
#endif

#define ReturnFailure(S) \
{ \
    char *__s = (char *)SS_Malloc(strlen(S)+1); \
    strcpy(__s, S); \
    return(__s); \
}

/*
 * start_secure_remote_node()
 *
 * Return: 0 on success
 *	   non-0 error code on failure
 */
static char *start_secure_remote_node(char *host,
				      int port,
				      char *remote_username,
				      char *dir,
				      char **env,
				      int n_env,
				      char *cmd,
				      char *args,
				      char *type)
{
    char myhost[256];
    int conn;
    struct passwd *passwd;
    char *local_username;
    char buf[250];
    int pid;
    int new_port, new_fd;
    int stdout_fd;
    struct sockaddr_in temp;
    int templen;
    int n;
    int rc;
    char msg[500];
    int i;

    myhost[0] = '\0';
    gethostname(myhost, sizeof(myhost));

    conn = connect_to_server(host, port);
    if (conn < 0)
    {
	nexus_dbgss_printf(NXDBG_SS_NODE, 1, ("start_secure_remote_node(): connect failed %s:%d\n", host, port));
	ReturnFailure("Failed to connect to server");
    }

    nexus_dbgss_printf(NXDBG_SS_NODE, 1,
		       ("start_secure_remote_node(): Connected\n"));

    passwd = getpwuid(geteuid());
    if (passwd == NULL)
    {
	local_username = getlogin();
	if (local_username == NULL)
	{
	    nexus_dbgss_printf(NXDBG_SS_NODE, 1,
			       ("No local username found\n"));
	    ReturnFailure("Failed to get local username from getpwuid() or getlogin()");
	}
    }
    else
    {
	local_username = passwd->pw_name;
    }

    if ((rc = send_string(conn, local_username)) != 0)
	ReturnFailure("Failed to send message to server");
    if (strcmp(remote_username, "USE_LOCAL_USERNAME") == 0)
    {
	if ((rc = send_string(conn, local_username)) != 0)
	    ReturnFailure("Failed to send message to server");
    }
    else
    {
        if ((rc = send_string(conn, remote_username)) != 0)
	    ReturnFailure("Failed to send message to server");
    }

#ifdef DONT_INCLUDE
    /*
     *  The comment was added because of the multi-threaded problem of a
     *  select being interrupted by another thread.  It should be enough
     *  to let the recv_string() figure out when the first message comes,
     *  but I'm keeping this here "just in case."
     */
    {
	struct timeval timeout;
	fd_set recv_fds;

	timeout.tv_sec = 1;
	timeout.tv_usec = 0;
	FD_ZERO(&recv_fds);
	FD_SET(conn, &recv_fds);
	n = select(conn+1, &recv_fds, NULL, NULL, &timeout);
	if (n == 1 || (n == -1 && errno == EINTR)) 
	{
	    if ((rc = recv_string(conn, buf, sizeof(buf))) != 0)
		ReturnFailure("Failed to receive message from server");
	}
	else
	{
	    if (n > 0)
	    {
		nexus_dbgss_printf(NXDBG_SS_NODE, 2, ("start_secure_remote_node(): wrong number of new connections: %d instead of 1\n", n));
		ReturnFailure("select() failed while waiting for message from server 1");
	    }
	    else
	    {
		nexus_dbgss_printf(NXDBG_SS_NODE, 2, ("start_secure_remote_node(): select() failed--errno = %d\n", errno));
		ReturnFailure("select() failed while waiting for message from server 2");
	    }
	}
    }
#else  /* DONT_INCLUDE */
    if ((rc = recv_string(conn, buf, sizeof(buf))) != 0)
	ReturnFailure("Failed to receive message from server");
#endif /* DONT_INCLUDE */

    nexus_dbgss_printf(NXDBG_SS_NODE, 2,
		       ("start_secure_remote_node(): Got reply1 %s\n", buf));

    if (strncmp(buf, "Password", 8) == 0)
    {
	if (strcmp(type, "normal") == 0)
	{
	    char pw_string[64];
	    getpw_ss(host, remote_username, pw_string, 64);
	    if ((rc = send_string(conn, pw_string)) != 0)
		ReturnFailure("Failed to send message to server");
	    if ((rc = recv_string(conn, buf, sizeof(buf))) != 0)
		ReturnFailure("Failed to receive message from server");
	}
	else if (strcmp(type, "iway") == 0)
	{
	    if ((rc = send_string(conn, remote_username)) != 0)
		ReturnFailure("Failed to send message to server");
	    if ((rc = recv_string(conn, buf, sizeof(buf))) != 0)
		ReturnFailure("Failed to receive message from server");
	}
	else
	{
	    ReturnFailure("Invalid ss_type value: must be \"normal\" or \"iway\"");
	}

	nexus_dbgss_printf(NXDBG_SS_NODE, 2,
			   ("start_secure_remote_node(): Got reply2 %s\n",
			    buf));

	if (strncmp(buf, "Proceed", 7) != 0)
	{
	    ReturnFailure(buf);
	}
    }
    else
    {
	if (strncmp(buf, "Proceed", 7) != 0)
	{
	    ReturnFailure(buf);
	}
    }

    nexus_dbgss_printf(NXDBG_SS_NODE, 2, ("start_secure_remote_node(): Validated on remote system\n"));

    if ((rc = send_string(conn, SERVER_CD_NOTIFIER)) != 0)
	ReturnFailure("Failed to send message to server");
    if ((rc = send_string(conn, dir)) != 0)
	ReturnFailure("Failed to send message to server");
    if ((rc = send_string(conn, SERVER_ENV_NOTIFIER)) != 0)
	ReturnFailure("Failed to send message to server");
    if ((rc = send_integer(conn, n_env)) != 0)
	ReturnFailure("Failed to send message to server");
    for (i = 0; i < n_env; i++)
    {
	if ((rc = send_integer(conn, strlen(env[i]))) != 0)
	    ReturnFailure("Failed to send message to server");
	if ((rc = send_string(conn, env[i])) != 0)
	    ReturnFailure("Failed to send message to server");
    }
    if ((rc = send_string(conn, cmd)) != 0)
	ReturnFailure("Failed to send message to server");
    if ((rc = send_string(conn, args)) != 0)
	ReturnFailure("Failed to send message to server");

    if ((pid = SS_Fork()) == 0)
    {
	setup_anon_listener(10, &new_port, &new_fd);
	send_integer(conn, new_port);
	fflush(stdout);
	templen = sizeof(temp);

	stdout_fd = accept(new_fd, (struct sockaddr *)&temp, &templen);
	close(new_fd);

	n = 1;
	while (n > 0)
	{
	    n = read(stdout_fd, msg, sizeof(msg)-1);
	    if (n > 0)
	    {
		msg[n] = '\0';
		nexus_stdio_lock();
		printf("%s\n", msg);
		nexus_stdio_unlock();
	    }
	}
#ifdef NEXUS_LISTENER_EXEC_HACK
	execl("/bin/true", "/bin/true", 0);
#endif
        exit(0);
    }

    if ((rc = recv_string(conn, buf, sizeof(buf))) != 0)
	ReturnFailure("Failed to receive message from server");

    nexus_dbgss_printf(NXDBG_SS_NODE, 2,
		       ("start_secure_remote_node(): Got reply3 %s\n", buf));

    if (strncmp(buf, "Success", 7) != 0)
    {
	kill(pid, SIGKILL);
	ReturnFailure(buf);
    }

    close(conn);

    return((char *) 0);
} /* start_secure_remote_node() */


/*
 * connect_to_server()
 */
static int connect_to_server(char *host, int server_port)
{
    int conn;
    struct hostent *hostent;
    struct sockaddr_in addr;
    int rc;
    int done;
    int save_errno;

    hostent = gethostbyname(host);

    addr.sin_family = hostent->h_addrtype;
    addr.sin_port = htons(server_port);
    memcpy(&addr.sin_addr, hostent->h_addr, hostent->h_length);

    done = NEXUS_FALSE;
    while (!done)
    {
	conn = socket(AF_INET, SOCK_STREAM, 0);
	if (conn < 0)
	{
	    /* socket() failed */
	    conn = -1;
	    done = NEXUS_TRUE;
	}

	if (connect(conn, (struct sockaddr *)&addr, sizeof(addr)) == 0)
	{
	    /* Success */
	    done = NEXUS_TRUE;
	}
	else
	{
	    save_errno = errno;
	    if (save_errno == EINPROGRESS)
	    {
		/*
		 * man connect: EINPROGRESS:
		 *   The socket is non-blocking and the connection
		 *   cannot be completed immediately. It is possible to
		 *   select(2) for completion by selecting the socket
		 *   for writing.
		 * So this connect has, for all practical purposes, succeeded.
		 */
		done = NEXUS_TRUE;
	    }
	    else if (   (save_errno == EINTR)
		     || (save_errno == ETIMEDOUT) )
	    {
		/*
		 * EINTR: This call got interrupted.
		 * ETIMEDOUT: The listener queue is probably full.
		 * So do nothing.  Just try again.
		 */
		close(conn);
	    }
	    else
	    {
		/* connect() failed */
		close(conn);
		conn = -2;
		done = NEXUS_TRUE;
	    }
	}
    }

    return(conn);
} /* connect_to_server() */


/*
 * send_integer()
 */
static int send_integer(int sock, int value)
{
    int rc;
    char str[10];

    sprintf(str, "%d", value);
    rc = send_string(sock, str);
    return (rc);
}


/*
 * send_string()
 */
static int send_string(int sock, char *str)
{
    int rc;
    int len = strlen(str);
    char nl = 10;

    rc = write(sock, str, len);
    if (rc < 0)
    {
	nexus_dbgss_printf(NXDBG_SS_NODE, 1, ("send_string():  write to secure server failed\n"));
	return(-3);
    }

    rc = write(sock, &nl, 1);
    if (rc < 0)
    {
	nexus_dbgss_printf(NXDBG_SS_NODE, 1, ("send_string():  write to secure server failed\n"));
	return(-4);
    }
    
    nexus_dbgss_printf(NXDBG_SS_NODE, 3,
		       ("send_string(): sent \"%s\" successfully\n", str));
    return(0);
} /* send_string() */


/*
 * recv_string()
 */
static int recv_string(int sock, char *buf, int len)
{
    char *bptr;
    int n;

    bptr = buf;
    while(1)
    {
	n = read(sock, bptr, 1);
	if (n < 0)
	{
	    nexus_dbgss_printf(NXDBG_SS_NODE, 1, ("recv_string(): read from secure server failed\n"));
	    return(-5);
	}

	if (*bptr == '\n') break;

	bptr++;
	if (bptr - buf >= len) break;
    }
    *bptr = 0;
    nexus_dbgss_printf(NXDBG_SS_NODE, 3,
		       ("recv_string(): read \"%s\" successfully\n", buf));
    return(0);
} /* recv_string() */


/*
 * setup_anon_listener()
 */
static int setup_anon_listener(int backlog, int *port, int *sock)
{
    int sin_len;
    struct sockaddr_in sin;
    int rc;

    *sock = socket(AF_INET, SOCK_STREAM, 0);
    if (*sock < 0)
    {
	nexus_dbgss_printf(NXDBG_SS_NODE, 1, ("setup_anon_listener(): unable to setup a listener socket\n"));
	return(-6);
    }

    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = htons(0);

    sin_len = sizeof(sin);

    rc = bind (*sock, (struct sockaddr *)&sin, sin_len);
    if (rc < 0)
    {
	nexus_dbgss_printf(NXDBG_SS_NODE, 1, ("setup_anon_listener(): unable to bind socket\n"));
	return(-7);
    }

    rc = listen (*sock, backlog);
    if (rc < 0)
    {
	nexus_dbgss_printf(NXDBG_SS_NODE, 1, ("setup_anon_listener(): unable to listen on to socket\n"));
	return(-8);
    }

    getsockname(*sock, (struct sockaddr *)&sin, &sin_len);
    *port = ntohs(sin.sin_port);

    return(0);

} /* setup_anon_listener() */


/*
 * getpw_ss()
 */
static void getpw_ss(char *host, char *username,
		     char *buf, int buf_length)
{
    char *s;

    nexus_stdio_lock();
    echo();
    printf("Password for %s@%s:  ", username, host);
    fflush(stdout);
    fgets(buf, buf_length-1, stdin);
    echo();
    printf("\n");
    nexus_stdio_unlock();

    for (s = buf; *s; s++)
    {
	if (*s == '\n')
	{
	    *s = '\0';
	    break;
	}
    }
} /* getpw_ss() */


#if defined(NEXUS_ARCH_NEXT040)

#include <sys/ioctl.h>

/*
 * echo()
 */
static int echo(void)
{
    struct sgttyb orig_tty, tty_new;

    if (ioctl(0, TIOCGETP, &orig_tty) < 0)
    {
	nexus_dbgss_printf(NXDBG_SS_NODE, 1, ("ioctl() TIOCGETP failed\n"));
	return(-9);
    }
    
    tty_new = orig_tty;
    tty_new.sg_flags &= ~(ECHO);

    if (ioctl(0, TIOCSETP, &tty_new) < 0)
    {
	nexus_dbgss_printf(NXDBG_SS_NODE, 1, ("ioctl() TIOSETP failed\n"));
	return(-10);
    }
    return(0);
} /* echo() */

#else

#include <termio.h>

/*
 * echo()
 */
static int echo(void)
{
    struct termio tty_orig, tty_new;

    if (ioctl(0, TCGETA, &tty_orig) < 0)
    {
	nexus_dbgss_printf(NXDBG_SS_NODE, 1, ("ioctl() TCGETA failed\n"));
	return(-11);
    }

    tty_new = tty_orig;
    tty_new.c_lflag &= ~(ECHO);

    if (ioctl(0, TCSETA, &tty_new) < 0)
    {
	nexus_dbgss_printf(NXDBG_SS_NODE, 1, ("ioctl() TCSETA failed\n"));
	return(-12);
    }

    return(0);
} /* echo() */

#endif
